/*
 * D-LITe : Copyright (c) 2011 Sylvain Cherrier
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 *$Id: dlite.c,v 0.9 2011/15/06 Sylvain Cherrier (sylvain.cherrier@univ-paris-est.fr)
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "contiki.h"
#include "contiki-net.h"
#include "rest.h"
#include "buffer.h"



#define WITH_BUTTON 1
#if WITH_BUTTON
    #include "dev/button-sensor.h"
#endif
//#include "dev/battery-sensor.h"
//#include "dev/sht11-sensor.h"
#include "dev/leds.h"

#define DEBUG 1
#if DEBUG
#define PRINTF(...) printf(__VA_ARGS__)
#define PRINT6ADDR(addr) PRINTF(" %02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x \n", ((u8_t *)addr)[0], ((u8_t *)addr)[1], ((u8_t *)addr)[2], ((u8_t *)addr)[3], ((u8_t *)addr)[4], ((u8_t *)addr)[5], ((u8_t *)addr)[6], ((u8_t *)addr)[7], ((u8_t *)addr)[8], ((u8_t *)addr)[9], ((u8_t *)addr)[10], ((u8_t *)addr)[11], ((u8_t *)addr)[12], ((u8_t *)addr)[13], ((u8_t *)addr)[14], ((u8_t *)addr)[15])
#define PRINTLLADDR(lladdr) PRINTF(" %02x:%02x:%02x:%02x:%02x:%02x ",(lladdr)->addr[0], (lladdr)->addr[1], (lladdr)->addr[2], (lladdr)->addr[3],(lladdr)->addr[4], (lladdr)->addr[5])
#else
#define PRINTF(...)
#define PRINT6ADDR(addr)
#define PRINTLLADDR(addr)
#endif




/* Pour IOTA*/

#define NB_RULES 50
#define WORD_SIZE 6
#define NB_OBSERVERS 20
char temp[100]; //buffer for answer
char rules[NB_RULES][4][WORD_SIZE]; //the set of rules
int rule_number = 0; //how many rules are in use
char state[WORD_SIZE]; //our state
uip_ipaddr_t observers[NB_OBSERVERS]; //nodes'list to notify
int nb_obser = 0; //how many observers

struct etimer et;
#define LOCAL_PORT 61618 
#define REMOTE_PORT 61616
char notify[WORD_SIZE+6]; //not really clean, but to transmit information throught process
int flag=0;

//functions i just need for address treatment  (strtoul is to heavy)

#define to_i(a) a>='a'?(a-'a'+10):(a-'0')



void action() { 	//link between IoTA and this node ! To be completed for each type of node !!!
	if (strcmp(state,"dark")==0) {
		leds_off(LEDS_BLUE);
	} else if (strcmp(state,"light")==0) {
			leds_on(LEDS_BLUE);
	}
	
}

static void my_coap_send_message(uip_ipaddr_t* obser_ipaddr) {
   if (init_buffer(200)) {
       //pour communication
    struct uip_udp_conn *client_conn_iota=NULL;
    int data_size = 0;
    int xact_id = 0;
 
       client_conn_iota = udp_new(obser_ipaddr, UIP_HTONS(REMOTE_PORT), NULL);
       if (client_conn_iota==NULL) {
           printf("ERROR\n");
           exit(-1);
       }
        udp_bind(client_conn_iota, UIP_HTONS(LOCAL_PORT));
 
#if DEBUG
       printf("send %s to ",notify);
        PRINT6ADDR(&client_conn_iota->ripaddr);
#endif
    

        coap_packet_t* attachment = (coap_packet_t*) allocate_buffer(sizeof (coap_packet_t));

        init_packet(attachment);
        coap_set_method(attachment, COAP_POST);
        attachment->tid = xact_id++;
        attachment->type = MESSAGE_TYPE_CON;
        coap_set_header_uri(attachment, "iota");
        attachment->payload = notify;
        attachment->payload_len = strlen(notify);


        data_size = serialize_packet(attachment, temp);
        uip_udp_packet_send(client_conn_iota, temp, data_size);

        delete_buffer();
        uip_udp_remove(client_conn_iota);
#if DEBUG
        printf("message send\n");
#endif
    }

}
void fst(char* message) {  //here we're trying to find what has to be done
                int i = 0;
                while ((i < rule_number) && (i < NB_RULES)) {
                    //printf("%d\n",i);
                    if (strncmp(message, rules[i][1], strlen(message)) == 0) { //rule found
                        //printf("rule found state %s and state %s\n", state, rules[i][0]);
                        if (strncmp(state, rules[i][0], strlen(state)) == 0) { //and good state
                            strncpy(temp, rules[i][3], WORD_SIZE);
                            strncpy(state, rules[i][2], WORD_SIZE); //new state
                            sprintf(temp, "state:%s", state); //to inform that we have changed
                            action(); //changing state.... checking if a real action is to be done...
                            sprintf(notify,"input=%s",rules[i][3]); // observers need to be warned !!
                            break; // rule has been found
                        }
                    }
                    i++;

                }

}

/* My Rules : my FST, and others commands*/
RESOURCE(myrules, METHOD_GET | METHOD_POST | METHOD_PUT | METHOD_DELETE, "iota");

void
myrules_handler(REQUEST* request, RESPONSE* response) {
    char input[WORD_SIZE]; //for Post, another node talks to us, ot a command for a FST (rules, link, list) (no more list)
    rest_set_response_status(response, OK_200); //default...

    //for debug,  i switch on and off green leds each time we have to deal with a command
    if (flag) leds_on(LEDS_GREEN); else leds_off(LEDS_GREEN);
    flag=1-flag;

    switch (coap_get_method(request)) {
        case COAP_GET:
            //strcpy(temp, state);
	    sprintf(temp,"name=TelosB\nmsg=push\nstate=dark\nstate=light"); //hardcode, as action and sensing
            break;
        case COAP_POST:
            if (coap_get_post_variable(request, "input", input, WORD_SIZE)) { //This message will be analyzed by our engine
#if DEBUG
                printf("One post me %s\n",input);
#endif

                sprintf(temp, "state:%s", state); //for information...
                fst(input);

            } else {
                rest_set_response_status(response, BAD_REQUEST_400);
                strcpy(temp, "no input");
            }

            break;
        case COAP_PUT: //to be called via cooper d1=value1&d2=value2&...
            if (coap_get_post_variable(request, "order", input, WORD_SIZE)) { //order analyzer
                if (strcmp(input, "rule") == 0) { //it's a new rule...
                    if (rule_number == NB_RULES) { //too many rules
                        rest_set_response_status(response, BAD_REQUEST_400);
                        strcpy(temp, "too many rules");
                    } else {
                        coap_get_post_variable(request, "state", rules[rule_number][0], WORD_SIZE);
                        coap_get_post_variable(request, "msg", rules[rule_number][1], WORD_SIZE);
                        coap_get_post_variable(request, "Nstate", rules[rule_number][2], WORD_SIZE);
                        coap_get_post_variable(request, "Smsg", rules[rule_number][3], WORD_SIZE);

                        sprintf(temp, " %d added", rule_number++);

                    }
                } else if (strcmp(input, "init") == 0) { //initial state
                    coap_get_post_variable(request, "state", state, WORD_SIZE);
                    sprintf(temp, "initial %s", state);
                } else if (strcmp(input, "link") == 0) { //we must store ipv6 address of sender : it want to be an observer
                    u16_t addr[8];
                    //address must be send in variables named a0 a1 a2 ... a7 (this will change in next version)
                    int i = 0;
                    for (i = 0; i < 8; i++) { //i need to get the 8 part of ipv6 address
                        char index[2] = "a*";
                        index[1] = '0' + i;
                        coap_get_post_variable(request, index, input, WORD_SIZE+1);
                          addr[i]=(((u16_t)to_i(input[0])))<<12;
                          addr[i]+=((u16_t)to_i(input[1]))<<8;
                          addr[i]+=((u16_t)to_i(input[2]))<<4;
                          addr[i]+=(to_i(input[3]));


#if DEBUG
                        printf("%s\n",input);
                        printf("is %u %u %u %u = %x \n",to_i(input[0]),
                                to_i(input[1]),to_i(input[2]),to_i(input[3]),addr[i]);
#endif
 
                    }


                    uip_ip6addr(&observers[nb_obser], addr[0], addr[1], addr[2], addr[3], addr[4], addr[5], addr[6], addr[7]);
#if DEBUG
                    printf("adding ");
                    PRINT6ADDR(&observers[nb_obser]);
#endif
                    nb_obser++;
                    sprintf(temp, "linked");


                } else {
                    rest_set_response_status(response, BAD_REQUEST_400);
                    sprintf(temp, "bad order : %s", input);

                }
            } else { //bad order
                rest_set_response_status(response, BAD_REQUEST_400);
                strcpy(temp, "no order ??");
            }


            break;
        case COAP_DELETE:  //cleaning values...
            strcpy(state,"");
            rule_number=0;
            nb_obser=0;

            strcpy(temp, "DEL");
            break;
    }


 
    rest_set_header_content_type(response, TEXT_PLAIN);
    rest_set_response_payload(response, temp, strlen(temp));


}


static void send_data() {
 //because our state has changed, we need to notify observers...
        int j;
        for (j = 0; j < nb_obser; j++) { //i need to notify all my observers
            my_coap_send_message(&observers[j]); //here we notify
#if DEBUG
            PRINT6ADDR(&observers[j]);
#endif
        }
   
}


RESOURCE(discover, METHOD_GET, ".well-known/core");

void
discover_handler(REQUEST* request, RESPONSE* response) {

    int index = 0;
    // index += sprintf(temp + index, "%s,", "</led>;n=\"Ledl\"");
    index += sprintf(temp + index, "%s", "</iota>;n=\"Me\"");

    rest_set_response_payload(response, temp, strlen(temp));
    rest_set_header_content_type(response, APPLICATION_LINK_FORMAT);
}

PROCESS(iota, "IoTA");
AUTOSTART_PROCESSES(&iota);


PROCESS_THREAD(iota, ev, data) {
    PROCESS_BEGIN();




    //SENSORS_ACTIVATE(light_sensor);
    
    rest_init();

    //rest_activate_resource(&resource_led);
    rest_activate_resource(&resource_myrules);
    rest_activate_resource(&resource_discover);


 
 #if WITH_BUTTON
  button_sensor.configure(SENSORS_ACTIVE, 1);
#endif
    etimer_set(&et, CLOCK_SECOND);
    while (1) {
        PROCESS_YIELD();
        if (etimer_expired(&et)) {
            if (strlen(notify)) {
                send_data();
                sprintf(notify, "");
            }
            etimer_reset(&et);
        }
#if WITH_BUTTON
        if(ev == sensors_event && data == &button_sensor) {
           uip_ipaddr_t myself;
           uip_ip6addr(&myself, 0, 0, 0, 0, 0, 0, 0, 0x1);
           fst("push");  //an action is like receiving a message
           printf("Button pressed \n");

        }
#endif
    }




    PROCESS_END();
}
